/*
 * Copyright (c) 2019 AlexIIL
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at https://mozilla.org/MPL/2.0/.
 */
package alexiil.mc.lib.multipart.api;

import java.util.function.Function;

import javax.annotation.Nullable;

import net.fabricmc.api.EnvType;
import net.fabricmc.api.Environment;
import net.minecraft.class_1058;
import net.minecraft.class_1059;
import net.minecraft.class_1109;
import net.minecraft.class_1268;
import net.minecraft.class_1269;
import net.minecraft.class_1657;
import net.minecraft.class_1792;
import net.minecraft.class_1799;
import net.minecraft.class_1922;
import net.minecraft.class_1937;
import net.minecraft.class_2246;
import net.minecraft.class_2248;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_2350.class_2352;
import net.minecraft.class_2371;
import net.minecraft.class_238;
import net.minecraft.class_2415;
import net.minecraft.class_243;
import net.minecraft.class_2470;
import net.minecraft.class_2487;
import net.minecraft.class_2498;
import net.minecraft.class_2586;
import net.minecraft.class_259;
import net.minecraft.class_265;
import net.minecraft.class_2680;
import net.minecraft.class_2960;
import net.minecraft.class_310;
import net.minecraft.class_3419;
import net.minecraft.class_3532;
import net.minecraft.class_3965;
import net.minecraft.class_47;
import net.minecraft.class_638;
import net.minecraft.class_702;
import net.minecraft.class_727;
import alexiil.mc.lib.net.IMsgReadCtx;
import alexiil.mc.lib.net.IMsgWriteCtx;
import alexiil.mc.lib.net.InvalidInputDataException;
import alexiil.mc.lib.net.NetByteBuf;
import alexiil.mc.lib.net.NetIdDataK;
import alexiil.mc.lib.net.NetIdDataK.IMsgDataWriterK;
import alexiil.mc.lib.net.NetIdSignalK;
import alexiil.mc.lib.net.NetIdTyped;
import alexiil.mc.lib.net.ParentNetIdSingle;

import alexiil.mc.lib.attributes.AttributeList;

import alexiil.mc.lib.multipart.api.MultipartContainer.MultipartCreator;
import alexiil.mc.lib.multipart.api.event.EventListener;
import alexiil.mc.lib.multipart.api.render.PartModelBaker;
import alexiil.mc.lib.multipart.api.render.PartModelKey;
import alexiil.mc.lib.multipart.impl.PartContainer;
import alexiil.mc.lib.multipart.impl.SingleReplacementBlockView;
import alexiil.mc.lib.multipart.impl.client.SingleSpriteProvider;
import alexiil.mc.lib.multipart.mixin.impl.BlockSoundGroupAccessor;

/** The base class for every part in a multipart block.
 * <p>
 * Generally implementations will want to override (in addition to the abstract methods):
 * <ul>
 * <li>{@link #getPickStack()} (and optionally {@link #addDrops(class_2371)}}</li>
 * </ul>
 */
public abstract class AbstractPart {

    public static final ParentNetIdSingle<AbstractPart> NET_ID;
    public static final NetIdDataK<AbstractPart> NET_RENDER_DATA;
    public static final NetIdSignalK<AbstractPart> NET_SPAWN_BREAK_PARTICLES;

    static {
        NET_ID = PartContainer.NET_KEY_PART;
        NET_RENDER_DATA = NET_ID.idData("render_data").toClientOnly()
            .setReadWrite(AbstractPart::readRenderData, AbstractPart::writeRenderData);
        // never buffer break particles because otherwise the block will be removed before the particles can spawn
        NET_SPAWN_BREAK_PARTICLES = NET_ID.idSignal("spawn_break_particles").toClientOnly().withoutBuffering()
            .setReceiver(AbstractPart::spawnBreakParticles);
    }

    public final PartDefinition definition;
    public final MultipartHolder holder;

    public AbstractPart(PartDefinition definition, MultipartHolder holder) {
        this.definition = definition;
        this.holder = holder;
    }

    public class_2487 toTag() {
        return new class_2487();
    }

    /** Writes the payload that will be passed into
     * {@link PartDefinition#loadFromBuffer(MultipartHolder, NetByteBuf, IMsgReadCtx)} on the client. (This is called on
     * the server and sent to the client). Note that this will be called *instead* of write and read payload. */
    public void writeCreationData(NetByteBuf buffer, IMsgWriteCtx ctx) {}

    public void writeRenderData(NetByteBuf buffer, IMsgWriteCtx ctx) {}

    public void readRenderData(NetByteBuf buffer, IMsgReadCtx ctx) throws InvalidInputDataException {}

    /** Sends the given {@link NetIdDataK} or {@link NetIdSignalK} to every player currently watching this multipart. */
    public final <T> void sendNetworkUpdate(T obj, NetIdTyped<T> netId) {
        holder.getContainer().sendNetworkUpdate(obj, netId);
    }

    /** Sends the given {@link NetIdDataK} to every player currently watching this multipart, with a custom
     * {@link IMsgDataWriterK}. */
    public final <T> void sendNetworkUpdate(T obj, NetIdDataK<T> netId, IMsgDataWriterK<T> writer) {
        holder.getContainer().sendNetworkUpdate(obj, netId, writer);
    }

    /** Sends the given {@link NetIdDataK} or {@link NetIdSignalK} to every player currently watching this multipart,
     * except for the given player. */
    public final <T> void sendNetworkUpdateExcept(@Nullable class_1657 except, T obj, NetIdTyped<T> netId) {
        holder.getContainer().sendNetworkUpdateExcept(except, obj, netId);
    }

    /** Sends the given {@link NetIdDataK} to every player currently watching this multipart, with a custom
     * {@link IMsgDataWriterK}, except for the given player. */
    public final <T> void sendNetworkUpdateExcept(
        @Nullable class_1657 except, T obj, NetIdDataK<T> netId, IMsgDataWriterK<T> writer
    ) {
        holder.getContainer().sendNetworkUpdateExcept(except, obj, netId, writer);
    }

    /** Called whenever this part was added to the {@link MultipartContainer}, either in
     * {@link class_2586#method_10996()} or when it is manually added by an item.
     * <p>
     * Register event handlers (as methods) with {@link MultipartEventBus#addListener(Object, Class, EventListener)}.
     * 
     * @param bus The event bus to register with. This is shorthand for
     *            <code>holder.getContainer().getEventBus()</code> */
    public void onAdded(MultipartEventBus bus) {
        // Nothing registers by default.
    }

    public void onRemoved() {
        // Nothing to do by default
    }

    /** Open method, that's designed to be called from the {@link class_1792} that places this part into the world.
     * <p>
     * (Nothing calls this by default, as all parts are placed from a custom item). */
    public void onPlacedBy(class_1657 player, class_1268 hand) {
        // Nothing to do by default
    }

    /** Called instead of {@link class_2248#method_9576(class_1937, class_2338, class_2680, class_1657)}, to play the broken sound,
     * and spawn break particles.
     * 
     * @return True if this should prevent {@link class_2248#method_9576(class_1937, class_2338, class_2680, class_1657)} from being
     *         called afterwards, false otherwise. */
    public boolean onBreak(class_1657 player) {
        if (!holder.getContainer().isClientWorld()) {
            playBreakSound();
            sendNetworkUpdate(this, NET_SPAWN_BREAK_PARTICLES);
        }
        return true;
    }

    /** @return The {@link class_2680} to use for {@link #playBreakSound()}, {@link #playHitSound(class_1657)},
     *         {@link #spawnBreakParticles()}, {@link #spawnHitParticle(class_2350)}, and
     *         {@link #calculateBreakingDelta(class_1657)}. */
    protected class_2680 getClosestBlockState() {
        return class_2246.field_10566.method_9564();
    }

    /** Called instead of {@link class_2248#method_9576(class_1937, class_2338, class_2680, class_1657)}, to play the broken sound,
     * and spawn break particles.
     * 
     * @return True if this should prevent {@link class_2248#method_9576(class_1937, class_2338, class_2680, class_1657)} from being
     *         called afterwards, false otherwise. */
    @Environment(EnvType.CLIENT)
    public void playHitSound(class_1657 player) {
        playHitSound(getClosestBlockState());
    }

    /** Called by default in {@link #onBreak(class_1657)} to play the breaking sound. The default implementation calls
     * {@link #playBreakSound(class_2680)} with {@link #getClosestBlockState()}. */
    protected void playBreakSound() {
        playBreakSound(getClosestBlockState());
    }

    protected final void playBreakSound(class_2680 blockState) {
        class_1937 world = holder.getContainer().getMultipartWorld();
        class_2498 group = blockState.method_26231();
        world.method_8396(
            null, holder.getContainer().getMultipartPos(),
            ((BlockSoundGroupAccessor) group).libmultipart_getBreakSound(), class_3419.field_15245,
            (group.method_10597() + 1.0F) / 2.0F, group.method_10599() * 0.8F
        );
    }

    @Environment(EnvType.CLIENT)
    protected final void playHitSound(class_2680 blockState) {
        class_2498 group = blockState.method_26231();
        class_310.method_1551().method_1483().method_4873(
            new class_1109(
                ((BlockSoundGroupAccessor) group).libmultipart_getHitSound(), class_3419.field_15245, //
                (group.method_10597() + 1.0F) / 8.0F, group.method_10599() * 0.8F, holder.getContainer().getMultipartPos()
            )
        );
    }

    private final void spawnBreakParticles(IMsgReadCtx ctx) throws InvalidInputDataException {
        ctx.assertClientSide();
        spawnBreakParticles();
    }

    /** Spawns a single breaking particle.
     * 
     * @param side The side that was hit
     * @return True to cancel the default breaking particle from spawning, false otherwise.
     * @deprecated This was renamed to {@link #spawnHitParticle(class_2350)} */
    @Environment(EnvType.CLIENT)
    @Deprecated
    public boolean spawnBreakingParticles(class_2350 side) {
        return spawnHitParticle(side);
    }

    /** Spawns a single partial-break (hit) particle.
     * 
     * @param side The side that was hit
     * @return True to cancel the default breaking particle from spawning, false otherwise. */
    @Environment(EnvType.CLIENT)
    public boolean spawnHitParticle(class_2350 side) {
        spawnHitParticle(side, getClosestBlockState());
        return true;
    }

    @Environment(EnvType.CLIENT)
    protected final void spawnHitParticle(class_2350 side, class_2680 state) {
        spawnHitParticle(side, state, (class_1058) null);
    }

    @Environment(EnvType.CLIENT)
    protected final void spawnHitParticle(class_2350 side, class_2680 state, @Nullable class_2960 spriteId) {
        class_1058 sprite;
        if (spriteId == null) {
            sprite = null;
        } else {
            sprite = getBlockAtlas().apply(spriteId);
        }
        spawnHitParticle(side, state, sprite);
    }

    @Environment(EnvType.CLIENT)
    protected final void spawnHitParticle(class_2350 side, class_2680 state, @Nullable class_1058 sprite) {
        spawnHitParticle(side, getOutlineShape().method_1107(), state, sprite);
    }

    @Environment(EnvType.CLIENT)
    protected final void spawnHitParticle(class_2350 side, class_238 box, class_2680 state, @Nullable class_1058 sprite) {
        class_1937 world = holder.getContainer().getMultipartWorld();
        class_2338 pos = holder.getContainer().getMultipartPos();
        class_702 manager = class_310.method_1551().field_1713;

        double x = pos.method_10263() + box.field_1323 + pos(world, side, class_2350.class_2351.field_11048, box.field_1320 - box.field_1323);
        double y = pos.method_10264() + box.field_1322 + pos(world, side, class_2350.class_2351.field_11052, box.field_1325 - box.field_1322);
        double z = pos.method_10260() + box.field_1321 + pos(world, side, class_2350.class_2351.field_11051, box.field_1324 - box.field_1321);

        class_727 particle = new class_727((class_638) world, x, y, z, 0, 0, 0, state, pos);
        particle.method_3075(0.2f);
        particle.method_3087(0.6f);
        if (sprite != null) {
            particle.method_18140(new SingleSpriteProvider(sprite));
        }
        manager.method_3058(particle);
    }

    protected static final double pos(class_1937 world, class_2350 side, class_2350.class_2351 axis, double size) {

        if (side.method_10166() == axis) {
            if (side.method_10171() == class_2352.field_11060) {
                return -0.1;
            } else {
                return size + 0.1;
            }
        }

        if (size >= 0.5) {
            return 0.1 + (size - 0.2) * world.field_9229.nextDouble();
        } else {
            double off = size / 8;
            return off + (size - 2 * off) * world.field_9229.nextDouble();
        }
    }

    private static Function<class_2960, class_1058> getBlockAtlas() {
        return class_310.method_1551().method_1549(class_1059.field_5275);
    }

    /** Called on the client to spawn break particles. This calls {@link #spawnBreakParticles(class_2680)} with
     * {@link #getClosestBlockState()} by default. */
    @Environment(EnvType.CLIENT)
    protected void spawnBreakParticles() {
        spawnBreakParticles(getClosestBlockState());
    }

    @Environment(EnvType.CLIENT)
    protected final void spawnBreakParticles(class_2680 state) {
        spawnBreakParticles(state, (class_1058) null);
    }

    @Environment(EnvType.CLIENT)
    protected final void spawnBreakParticles(class_2680 state, @Nullable class_2960 spriteId) {
        class_1058 sprite;
        if (spriteId == null) {
            sprite = null;
        } else {
            sprite = getBlockAtlas().apply(spriteId);
        }
        spawnBreakParticles(state, sprite);
    }

    @Environment(EnvType.CLIENT)
    protected final void spawnBreakParticles(class_2680 state, @Nullable class_1058 sprite) {
        class_1937 world = holder.getContainer().getMultipartWorld();
        class_2338 pos = holder.getContainer().getMultipartPos();
        class_702 manager = class_310.method_1551().field_1713;
        class_265 voxelShape = getOutlineShape();
        for (class_238 box : voxelShape.method_1090()) {
            double x0 = box.field_1323;
            double y0 = box.field_1322;
            double z0 = box.field_1321;
            double x1 = box.field_1320;
            double y1 = box.field_1325;
            double z1 = box.field_1324;
            double minX = Math.min(1.0D, x1 - x0);
            double minY = Math.min(1.0D, y1 - y0);
            double minZ = Math.min(1.0D, z1 - z0);
            int maxX = Math.max(2, class_3532.method_15384(minX / 0.25D));
            int maxY = Math.max(2, class_3532.method_15384(minY / 0.25D));
            int maxZ = Math.max(2, class_3532.method_15384(minZ / 0.25D));

            for (int x = 0; x < maxX; ++x) {
                for (int y = 0; y < maxY; ++y) {
                    for (int z = 0; z < maxZ; ++z) {
                        double vX = (x + 0.5D) / maxX;
                        double vY = (y + 0.5D) / maxY;
                        double vZ = (z + 0.5D) / maxZ;
                        double pX = vX * minX + x0;
                        double pY = vY * minY + y0;
                        double pZ = vZ * minZ + z0;
                        class_727 particle = new class_727(
                            (class_638) world, pos.method_10263() + pX, pos.method_10264() + pY, pos.method_10260() + pZ, vX - 0.5D,
                            vY - 0.5D, vZ - 0.5D, state, pos
                        );
                        if (sprite != null) {
                            particle.method_18140(new SingleSpriteProvider(sprite));
                        }
                        manager.method_3058(particle);
                    }
                }
            }
        }
    }

    protected final void addRequiredPart(AbstractPart required) {
        holder.addRequiredPart(required);
    }

    protected final void removeRequiredPart(AbstractPart required) {
        holder.removeRequiredPart(required);
    }

    /** @return The {@link class_265} to use for calculating if this pluggable overlaps with another pluggable. */
    public abstract class_265 getShape();

    /** Checks to see if this {@link AbstractPart} can overlap with the other part. Note that this is only called if the
     * {@link #getShape()} of this part intersects with the {@link #getShape()} of the other part. However this is never
     * called if the shape of one of the parts is completely contained by the shape of the other part.
     * <p>
     * This is called once for each part currently contained in a {@link MultipartContainer} in
     * {@link MultipartContainer#offerNewPart(MultipartCreator, boolean)}. */
    public boolean canOverlapWith(AbstractPart other) {
        return false;
    }

    /** @return The shape to use when solidity logic, collisions with entities, ray tracing, etc. This should always
     *         encompass {@link #getShape()}. */
    public class_265 getCollisionShape() {
        return getShape();
    }

    /** @return The shape to use when calculating lighting and checking for opacity. This may be empty, although it
     *         should always be contained by {@link #getCollisionShape()}. Generally anything that's not opaque should
     *         override this and return {@link class_259#method_1073()}. */
    public class_265 getCullingShape() {
        return getCollisionShape();
    }

    /** @return The shape for rendering bounding boxes and ray tracing. */
    public class_265 getOutlineShape() {
        return getCollisionShape();
    }

    /** @return The (potentially dynamic) shape for rendering bounding boxes and ray tracing. Unlike
     *         {@link #getOutlineShape()} this is only called when rendering the box for this specific part. */
    public class_265 getDynamicShape(float partialTicks) {
        return getOutlineShape();
    }

    /** @return True if this pluggable should be an {@link AttributeList#obstruct(class_265) obstacle} for attributes
     *         with it's {@link #getShape()} when searching in this particular direction. */
    public boolean isBlocking(class_2350 searchDirection) {
        return true;
    }

    /** Offers every contained attribute to the given attribute list. NOTE: This must always use
     * {@link AttributeList#offer(Object, class_265)} with {@link #getShape()} as the {@link class_265} argument!
     * <p>
     * The default implementation will {@link AttributeList#obstruct(class_265)} the {@link #getShape()} if
     * {@link #isBlocking(class_2350)} returns true, and the search direction is not null. */
    public void addAllAttributes(AttributeList<?> list) {
        class_2350 searchDirection = list.getSearchDirection();
        if (searchDirection != null && isBlocking(searchDirection)) {
            list.obstruct(getShape());
        }
    }

    /** Called whenever this part is picked by the player (similar to
     * {@link class_2248#method_9574(class_1922, class_2338, class_2680)})
     * 
     * @return The stack that should be picked, or ItemStack.EMPTY if no stack can be picked from this part. */
    public class_1799 getPickStack() {
        return class_1799.field_8037;
    }

    /** A target for retrieving item drops from {@link AbstractPart#addDrops(ItemDropTarget, class_47)}. This will
     * either spawn the dropped items into the world, or add them to a list, depending on what method called it. */
    public interface ItemDropTarget {

        /** Drops the given itemstack, using a default position and velocity. */
        void drop(class_1799 stack);

        default void dropAll(Iterable<class_1799> iter) {
            for (class_1799 stack : iter) {
                drop(stack);
            }
        }

        /** Drops the given {@link class_1799}, using the given position and a random velocity. This will also split the
         * stack up randomly, like how vanilla splits stacks. */
        void drop(class_1799 stack, class_243 pos);

        /** If {@link #dropsAsEntity()} is false then this ignores the position and velocity. Otherwise, this drops the
         * given {@link class_1799} directly at the given position, using the exact given velocity. This won't split the
         * stack up. */
        void drop(class_1799 stack, class_243 pos, class_243 velocity);

        /** @return True if the position and velocity of the {@link class_1799} will be used. */
        boolean dropsAsEntity();
    }

    public void addDrops(ItemDropTarget target, class_47 context) {
        class_2371<class_1799> list = class_2371.method_10211();
        addDrops(list);
        if (!list.isEmpty()) {
            target.dropAll(list);
        }
    }

    /** @deprecated Replaced by {@link #addDrops(ItemDropTarget, class_47)} */
    @Deprecated
    public void addDrops(class_2371<class_1799> to) {
        class_1799 pickStack = getPickStack();
        if (!pickStack.method_7960()) {
            to.add(pickStack);
        }
    }

    /** Called instead of {@link class_2248#method_9556(class_1937, class_1657, class_2338, class_2680, class_2586, class_1799)},
     * except that this shouldn't drop any items, as that's handled separately. */
    public void afterBreak(class_1657 player) {
        player.method_7322(0.005F);
    }

    /** Part version of {@link class_2248#method_9594(class_2680, class_1657, class_1922, class_2338)}.
     * <p>
     * The default implementation treats parts as equal to {@link #getClosestBlockState()}. */
    public float calculateBreakingDelta(class_1657 player) {
        return calculateBreakingDelta(player, getClosestBlockState());
    }

    /** Calculates {@link #calculateBreakingDelta(class_1657)} as if this part was the given block instead. */
    public final float calculateBreakingDelta(class_1657 player, class_2248 block) {
        return calculateBreakingDelta(player, block.method_9564());
    }

    /** Calculates {@link #calculateBreakingDelta(class_1657)} as if this part was the given block state instead. */
    public final float calculateBreakingDelta(class_1657 player, class_2680 state) {
        class_1937 world = holder.getContainer().getMultipartWorld();
        class_2338 thisPos = holder.getContainer().getMultipartPos();
        float hardness = state.method_26214(new SingleReplacementBlockView(world, thisPos, state), thisPos);
        return calcBreakingDelta(player, state, hardness);
    }

    /** Calculates {@link #calculateBreakingDelta(class_1657)} as if this part was the given block state instead, but
     * using a custom hardness value. */
    public static float calcBreakingDelta(class_1657 player, class_2680 state, float hardness) {
        if (hardness == -1.0F) {
            return 0.0F;
        } else {
            int mult = player.method_7305(state) ? 30 : 100;
            return player.method_7351(state) / hardness / mult;
        }
    }

    /** Called whenever this part is used via
     * {@link class_2248#method_9534(class_2680, class_1937, class_2338, class_1657, class_1268, class_3965)}. */
    public class_1269 onUse(class_1657 player, class_1268 hand, class_3965 hit) {
        return class_1269.field_5811;
    }

    /** Called whenever {@link class_2586#applyRotation(BlockRotation)} is called on the containing block.
     * 
     * @param rotation A rotation. LMP never calls this with {@link class_2470#field_11467} */
    public void rotate(class_2470 rotation) {

    }

    /** Called whenever {@link class_2586#applyMirror(BlockMirror)} is called on the containing block.
     * 
     * @param mirror A mirror. LMP never calls this with {@link class_2415#field_11302} */
    public void mirror(class_2415 mirror) {

    }

    /** Called on the client for both rendering, and checking if this needs to re-render in
     * {@link MultipartContainer#redrawIfChanged()}.
     * <p>
     * This is no longer called on the server.
     * 
     * @return The {@link PartModelKey} for the {@link PartModelBaker} to use to emit a static model. Returning null
     *         will bake nothing. */
    @Nullable
    public abstract PartModelKey getModelKey();
}
